const APPNAME: &'static str = "Source Cubemapper (Server)";

extern crate clap;

use clap::{App, Arg};
use interprocess::local_socket::{LocalSocketListener, LocalSocketStream};
use bincode::{serialize, deserialize};
use source_cubemapper::*;
use std::{
    ffi::OsStr,
    io::{self, BufRead, BufReader, Write},
    process::Command,
    time::Instant,
};

struct GameOptions<'a> {
    path: &'a OsStr,
    game: Option<&'a str>,
}

fn handle_error(connection: io::Result<LocalSocketStream>) -> Option<LocalSocketStream> {
    connection
        .map_err(|error| eprintln!("Incoming connection failed: {}", error))
        .ok()
}

fn main() -> Result<()> {
    /////////////////////////////////////////////////////////////////////////
    // CLI
    let long_version = format!(
        "\
        \nRevision: ({})\
        \nDate:     {}\
        \nVersion:  {}\
        \nBuilt:    {}",
        REVISION, REVDATE, VERSION, BUILDTD,
    );

    let matches = App::new(APPNAME)
        .version(&*REVISION)
        .long_version(&*long_version)
        .author(AUTHORS)
        .long_about(format!(
            "{}\n \
            Awaits jobs for building cubemaps, launches the game, and notifies the client of results.",
            APPNAME
        ).as_str())
        .arg(
            Arg::with_name("path")
                .long("path")
                .help("Path to source executable")
                .takes_value(true)
                .required(true),
        )
        .arg(
            Arg::with_name("game")
                .help("Game launch with, e.g: cstrike")
                .long("game")
                .takes_value(true),
        )
        .get_matches();
    // CLI
    /////////////////////////////////////////////////////////////////////////
    let opts = GameOptions {
        path: matches
            .value_of_os("path")
            .expect("Couldn't get game path!"),
        game: matches.value_of("game"),
    };
    loop {
        let listener =
            LocalSocketListener::bind(SOCK_LOCATION).expect("Could not bind to local socket!");
        println!(
            "{}
            Waiting for Jobs at: {}\n",
            APPNAME, SOCK_LOCATION,
        );
        for mut conn in listener.incoming().filter_map(handle_error) {
            println!("Client connected!");
            //conn.write_all(b"Hello from server!\n")?;

            // Add buffering to the connection to read a line.
            let mut incoming = BufReader::new(&mut conn);
            let mut buffer = Vec::new();
            incoming.read_until(0xA, &mut buffer)?;
            buffer.pop();
            //println!("{:#?}", buffer);
            let deser = deserialize::<Message>(buffer.as_ref());
            match deser {
                Ok(v) => match v {
                    Message::Job(j) => {
                        println!(
                            "Job recieved:\n\
                                \tMap: {}\n\
                                \tHDR/LDR {}/{}",
                            j.map, j.dnr.HDR, j.dnr.LDR,
                        );
                        let mut ser = serialize(&Message::Ack(true)).unwrap();
                        ser.push(0xA);
                        conn.write_all(&ser).unwrap();
                        {
                            let result = launch_cubemap_job(&j, &opts);
                            match result {
                                Ok(_) => {
                                    let result = Message::Success((0, "Job Complete".to_string()));
                                    let mut ser = serialize(&result).unwrap();
                                    ser.push(0xA);
                                    conn.write_all(&ser).unwrap();
                                }
                                Err(e) => {
                                    let code = match e {
                                        Error::BadEngineExit(v) => v,
                                        _ => 1,
                                    };
                                    let message = format!("{:?}", e);
                                    let result = Message::Failure((code, message));
                                    let mut ser = serialize(&result).unwrap();
                                    ser.push(0xA);
                                    conn.write_all(&ser).unwrap();
                                }
                            }
                        }
                        continue;
                    }
                    _ => {
                        eprintln!("Attempted connection start with non job!");
                        let mut ser = serialize(&Message::Ack(false)).unwrap();
                        ser.push(0xA);
                        conn.write_all(&ser).unwrap();
                        continue;
                    }
                },
                Err(e) => {
                    eprintln!("Couldn't deserialize for job! {}", e);
                    let mut ser = serialize(&Message::Ack(false)).unwrap();
                    ser.push(0xA);
                    conn.write_all(&ser).unwrap();
                    continue;
                }
            }
        }
    }
    println!("Exiting.");
    Ok(())
}

fn launch_cubemap_job(job: &Job, opts: &GameOptions) -> Result<()> {
    let start_time = Instant::now();
    let map = &job.map;
    // Need to run this whole thing for LDR and HDR if both are enabled.
    for n in 0..2 {
        let (hdr_level, hdr_name) = match n {
            0 => {
                match job.dnr.HDR {
                    true => (2, "HDR"),
                    false => continue, // skip this pass!
                }
            }
            1 => match job.dnr.LDR {
                true => (0, "LDR"),
                false => continue, // skip this pass!
            },
            _ => break,
        };
        println!("Building {} for {}!", hdr_name, map);
        let mut process = Command::new(opts.path);
        if opts.game.is_some() {
            process.args(&["-game", opts.game.unwrap()]);
        }
        process
            .arg("-novid")
            .arg("-dev")
            .arg("-condebug")
            .arg("-hijack")
            .arg("-conclearlog")
            .args(&["+sv_cheats", "1"])
            .args(&["+building_cubemaps", "1"])
            .args(&["+mat_hdr_level", hdr_level.to_string().as_str()])
            .args(&["+mat_specular", "0"])
            .args(&["+r_lod", "0"])
            .args(&["+map", format!("{}", map).as_str()])
            // We use this mess because -buildcubemaps doesn't work for LDR.
            // This has the benefit of ensuring the game closes even if the map doesn't exist.
            .args(&["+wait", "30; buildcubemaps; wait 30; exit"]);

        let out = process.spawn()?.wait()?;
        if out.success() != true {
            eprintln!("Job failed!");
            eprintln!("{:?}", out);
            return Err(Error::BadEngineExit(out.code().unwrap_or(1)));
        }
    }
    let dur = (Instant::now() - start_time).as_secs_f32();
    println!("Job completed in {} seconds", dur);
    Ok(())
}
